#!/usr/bin/env python3

"""
Goes through a Sprint Bliss XML, and dumps all the orthography on stdout.
"""

from __future__ import annotations

from argparse import ArgumentParser

import _setup_returnn_env  # noqa

# noinspection PyProtectedMember
from returnn.datasets.lm import _iter_bliss


def main():
    """
    Main entry.
    """
    parser = ArgumentParser(description="dump orth from Bliss XML file as-is")
    parser.add_argument("xml")
    args = parser.parse_args()
    corpus_filename = args.xml

    def callback(orth):
        """
        :param str orth:
        """
        print(orth)

    _iter_bliss(filename=corpus_filename, callback=callback)


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    main()



"""
Data iterator on a data loader which an additional queue
"""

from __future__ import annotations

from typing import Union, TypeVar, Iterator, Iterable, Sequence

T = TypeVar("T")


class QueuedDataIter:
    """
    Wraps another iterator, and also has a queue which allows to put items back in.
    """

    def __init__(self, original_iter: Union[Iterable[T], Iterator[T]]):
        self._iter = iter(original_iter)
        self._queue = []

    def extend(self, items: Sequence[T]):
        """
        Extend the queue. The next items by :func:`__next__` will be the ones from ``items``, in the same order.
        """
        # reversed because we pop from behind.
        self._queue.extend(reversed(items))

    def __iter__(self):
        return self

    def __next__(self) -> T:
        if self._queue:
            return self._queue.pop(-1)
        return next(self._iter)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import _setup_test_env  # noqa
import numpy as np
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_pack_padded():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, pack_dim = rf.pack_padded(x, dims=[batch_dim, time_dim], enforce_sorted=False)
            return pack, pack_dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, pack_dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(pack_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_masked_select():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "mask": Tensor("mask", [batch_dim, time_dim], dtype="bool"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        mask = extern_data["mask"]
        out, pack_dim = rf.masked_select(x, mask=mask, dims=[batch_dim, time_dim])
        out.mark_as_default_output(shape=(pack_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_masked_select_single_dim():
    # https://github.com/rwth-i6/returnn/issues/1605
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "mask": Tensor("mask", [batch_dim, time_dim], dtype="bool"),
        }
    )

    # noinspection PyShadowingNames
    def _forward_step(*, extern_data: TensorDict, **_kwargs):
        x = extern_data["data"]
        mask = extern_data["mask"]
        out, pack_dim = rf.masked_select(x, mask=mask, dims=[time_dim])
        out.mark_as_default_output(shape=(batch_dim, pack_dim, in_dim))
        pack_dim.dyn_size_ext.mark_as_output("out_size", shape=[batch_dim])
        x.mark_as_output("input", shape=[batch_dim, time_dim, in_dim])
        mask.mark_as_output("mask", shape=[batch_dim, time_dim])
        time_dim.dyn_size_ext.mark_as_output("in_size", shape=[batch_dim])

    res = run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    for k, v in res.data.items():
        print(f"{k} {v} = {v.raw_tensor}")
    num_batches = res["input"].raw_tensor.shape[0]
    have_non_zero = False
    have_less = False
    for b in range(num_batches):
        in_size = res["in_size"].raw_tensor[b]
        out_size = res["out_size"].raw_tensor[b]
        print(f"batch {b}, in_size {in_size}, out_size {out_size}")
        assert 0 <= out_size <= in_size
        if out_size > 0:
            have_non_zero = True
        if out_size < in_size:
            have_less = True
        c = 0
        for t in range(in_size):
            if res["mask"].raw_tensor[b, t]:
                c += 1
        assert c == out_size
        t_ = 0
        for t in range(in_size):
            if res["mask"].raw_tensor[b, t]:
                assert (res["input"].raw_tensor[b, t] == res["output"].raw_tensor[b, t_]).all()
                t_ += 1
    assert have_non_zero and have_less  # just that the test case covered all cases


def test_reshape():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reshape(x, in_dims=(time_dim, in_dim), out_dims=(in_dim, time_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_expand_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    expand_dim = Dim(3, name="expand")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.expand_dim(x, expand_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, expand_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_concat():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            return rf.concat((x, in_dim), (x, in_dim))

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, dim = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim, Dim]]:
            pack, (new_time, new_feat) = rf.pad(x, axes=[time_dim, in_dim], padding=[(1, 2), (3, 4)], value=0)
            return pack, (new_time, new_feat)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time, new_feat) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, new_feat))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad_time():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim]]:
            pack, (new_time,) = rf.pad(x, axes=[time_dim], padding=[(1, 0)], value=0)
            return pack, (new_time,)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, (new_time,) = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_pad_time_right():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Tuple[Dim]]:
            pack, (new_time,) = rf.pad(x, axes=[time_dim], padding=[(0, 1)], value=1)
            return pack, (new_time,)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        data = extern_data["data"]
        data.mark_as_output("data", shape=(batch_dim, time_dim, in_dim))
        out, (new_time,) = model(data)
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    res = run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)
    data_: Tensor = res["data"]
    out_: Tensor = res["output"]
    assert data_.dims == (batch_dim, time_dim, in_dim)
    new_time_dim = out_.dims[1]
    assert out_.dims == (batch_dim, new_time_dim, in_dim) and new_time_dim != time_dim
    assert new_time_dim == time_dim + 1  # math dim... not really necessary check here...
    assert time_dim.dyn_size_ext.dims == new_time_dim.dyn_size_ext.dims == (batch_dim,)
    batch_size = batch_dim.get_dim_value()
    assert batch_size > 1
    assert len(set(time_dim.dyn_size_ext.raw_tensor)) > 1  # not all the same
    for b in range(batch_size):
        seq_len = time_dim.dyn_size_ext.raw_tensor[b]
        new_seq_len = new_time_dim.dyn_size_ext.raw_tensor[b]
        print(f"batch {b}, seq_len {seq_len}, new_seq_len {new_seq_len}")
        assert new_seq_len == seq_len + 1
        np.testing.assert_allclose(data_.raw_tensor[b, :seq_len], out_.raw_tensor[b, :seq_len])
        print(out_.raw_tensor[b])
        assert all(out_.raw_tensor[b, seq_len] == 1.0)


def test_stack():
    batch_dim_ = Dim(3, name="batch")
    time_dim = Dim(5, name="time")

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        seq = rf.range_over_dim(time_dim)  # [T]
        out, _ = rf.stack([seq, seq, seq], out_dim=batch_dim_)  # [B,T]
        out.mark_as_default_output(shape=(batch_dim_, time_dim))

    out = run_model(TensorDict(), lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    out = out["output"]
    assert out.dims == (batch_dim_, time_dim)
    assert out.raw_tensor.tolist() == [[0, 1, 2, 3, 4], [0, 1, 2, 3, 4], [0, 1, 2, 3, 4]]


def test_gather():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.gather(x, indices=0, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_2d_indices():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
            "classes": Tensor("classes", [batch_dim, time_dim], dtype="int32", sparse_dim=in_dim),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor, y: Tensor) -> Tensor:
            return rf.gather(x, indices=y, axis=in_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], extern_data["classes"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_gather_feature_dim():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], feature_dim=in_dim, dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.gather(x, indices=0, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        assert out.feature_dim == in_dim
        out.mark_as_default_output(shape=(batch_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_slice():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tuple[Tensor, Dim]:
            pack, new_time = rf.slice(x, axis=time_dim, start=1, size=2)
            return pack, new_time

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, new_time = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, new_time, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_shift_right():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim], sparse_dim=in_dim, dtype="int32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.shift_right(x, axis=time_dim, pad_value=0)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_reverse_sequence():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor) -> Tensor:
            return rf.reverse_sequence(x, axis=time_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_where():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "cond": Tensor("cond", [batch_dim, time_dim], dtype="bool"),
            "true": Tensor("true", [batch_dim, time_dim, in_dim], dtype="float32"),
            "false": Tensor("false", [batch_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.where(extern_data["cond"], extern_data["true"], extern_data["false"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_search_sorted():
    batch_dim_ = Dim(3, name="batch")
    time_dim = Dim(13, name="time")

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        sorted_seq = rf.range_over_dim(time_dim, dtype="float32")  # [T]
        index1 = rf.search_sorted(sorted_seq, rf.constant(4.5, dims=()), axis=time_dim)  # [] -> T
        assert index1.dims == () and index1.sparse_dim == time_dim and index1.dtype == "int32"
        index1.mark_as_output("index1", shape=())
        index2 = rf.search_sorted(sorted_seq, rf.constant(4.0, dims=()), axis=time_dim)  # [] -> T
        assert index2.dims == () and index2.sparse_dim == time_dim and index2.dtype == "int32"
        index2.mark_as_output("index2", shape=())
        index3 = rf.search_sorted(sorted_seq, rf.constant(4.0, dims=()), axis=time_dim, side="right")  # [] -> T
        assert index3.dims == () and index3.sparse_dim == time_dim and index3.dtype == "int32"
        index3.mark_as_output("index3", shape=())
        index4 = rf.search_sorted(
            sorted_seq, rf.convert_to_tensor(np.array([4.0, 4.5, 5.0]), dims=[batch_dim_]), axis=time_dim
        )  # [] -> T
        assert index4.dims == (batch_dim_,) and index4.sparse_dim == time_dim and index4.dtype == "int32"
        index4.mark_as_output("index4", shape=(batch_dim_,))

    out = run_model(TensorDict(), lambda *, epoch, step: rf.Module(), _forward_step, test_tensorflow=False)
    assert out["index1"].dims == () and out["index1"].raw_tensor.item() == 5
    assert out["index2"].dims == () and out["index2"].raw_tensor.item() == 4
    assert out["index3"].dims == () and out["index3"].raw_tensor.item() == 5
    assert out["index4"].dims == (batch_dim_,) and out["index4"].raw_tensor.tolist() == [4, 5, 5]


def test_where_int():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "cond": Tensor("cond", [batch_dim, time_dim], dtype="bool"),
            "true": Tensor("true", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Module, extern_data: TensorDict):
        out = rf.where(extern_data["cond"], extern_data["true"], 0)
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_copy_masked():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Conv1d, extern_data: TensorDict):
        x = extern_data["data"]
        x = x.copy_masked(1)
        # Do some pooling to make sure the copy_masked has an effect on the output.
        x, _ = rf.pool1d(x, mode="avg", pool_size=3, strides=1, padding="same", in_spatial_dim=time_dim)
        x.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)


def test_cast_sparse():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    # noinspection PyShadowingNames,PyUnusedLocal
    def _forward_step(*, model: rf.Conv1d, extern_data: TensorDict):
        x = rf.reduce_argmax(extern_data["data"], axis=in_dim)
        assert x.sparse_dim == in_dim
        x.mark_as_output("argmax", shape=[batch_dim, time_dim])
        rf.cast(x, "float32").mark_as_output("float", shape=[batch_dim, time_dim])

    run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)



"""
Internal utils
"""

from __future__ import annotations
from typing import Union, Optional, Any, Type, TypeVar, Sequence, Tuple, List
import numpy

from returnn import frontend as _global_rf
from returnn.frontend._backend import Backend, get_backend_by_raw_tensor_type
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf

T = TypeVar("T")


def get_backend_from_tensors(*args):
    """
    :param args:
    :return: frontend, fallback to global frontend
    """
    for x in args:
        if isinstance(x, Tensor):
            # noinspection PyProtectedMember
            return x._raw_backend
    return _global_rf


def get_dtype_name(x: Union[T, Tensor[T], int, float]) -> str:
    """
    :param x: tensor
    :return: dtype of tensor, as string
    """
    if isinstance(x, Tensor):
        return x.dtype
    elif isinstance(x, int):
        return rf.get_default_int_dtype()
    elif isinstance(x, float):
        return rf.get_default_float_dtype()
    else:
        backend = get_backend_by_raw_tensor_type(type(x))
        return backend.get_dtype_name_raw(x)


def is_int(x: Union[T, Tensor[T], int, float]) -> bool:
    """
    :param x:
    :return: whether the dtype is int
    """
    dtype = get_dtype_name(x)
    return dtype.startswith("int") or dtype.startswith("uint")


# There is a native implementation of this (_native tensorCopyTemplateSimple, compareAndCombine).
def bin_op_out_template(
    backend: Type[Backend],
    a: Union[Tensor[T], int, float, numpy.number],
    b: Union[Tensor[T], int, float, numpy.number],
    *,
    name: str,
    copy_sparse_dim: bool = True,
    allow_broadcast_all_sources: Optional[bool] = None,
    dim_order: Optional[Sequence[Dim]] = None,
    allow_scalar: bool = True,
) -> Tuple[Tensor[T], T, T]:
    """
    make template for output tensor of binary op

    :param backend:
    :param a:
    :param b:
    :param name: for returned Tensor. no other functionality
    :param copy_sparse_dim:
    :param allow_broadcast_all_sources: if True, it is allowed that neither a nor b has all dims of the result.
        Not needed when out_dims is specified explicitly.
    :param dim_order: defines the order of the resulting dims. if None, it is automatically inferred from a and b.
        Not all the dims of a and b need to be specified here, and there could also be other dims in the dim_order.
    :param allow_scalar: if True, it is allowed that a or b is a scalar, and then no broadcast dims are added.
        This can be relevant to allow things like x * 2, where x in on GPU, and then PyTorch allows 2 to stay on CPU.
    :return: out, a_raw, b_raw
    """
    src_dtype = None
    src_device = None
    if isinstance(a, Tensor):
        src_dtype = a.dtype
        src_device = a.device
    elif isinstance(b, Tensor):
        src_dtype = b.dtype
        src_device = b.device
    a = rf.convert_to_tensor(a, dtype=src_dtype, device=src_device, keep_scalar_on_cpu=allow_scalar, _backend=backend)
    src_dtype = src_dtype or a.dtype
    b = rf.convert_to_tensor(b, dtype=src_dtype, device=src_device, keep_scalar_on_cpu=allow_scalar, _backend=backend)
    # sanity checks
    # noinspection PyProtectedMember
    assert a._raw_backend == b._raw_backend, "Cannot combine tensors from two different frontends, e.g. TF and PT"
    all_dims = []
    for dim in a.dims + b.dims:
        if dim in all_dims:
            continue
        # Not simply `all_dims.append(dim)`,
        # because a dim might occur multiple times in a.dims or b.dims
        # (with different match_priority),
        # e.g. in the case of square matrices.
        # Still it is the common case that they are unique,
        # and this allows for a faster path.
        if a.dims.count(dim) <= 1 and b.dims.count(dim) <= 1:
            all_dims.append(dim)
            continue
        if a.dims.count(dim) >= b.dims.count(dim):
            all_dims.extend([dim_ for dim_ in a.dims if dim_ == dim])
        else:
            all_dims.extend([dim_ for dim_ in b.dims if dim_ == dim])
    if all(set(x.dims) != set(all_dims) for x in (a, b)):
        if allow_broadcast_all_sources is False:
            raise ValueError(f"compare: sources {a!r} {b!r} not allowed with allow_broadcast_all_sources=False")
        elif allow_broadcast_all_sources is None:
            raise ValueError(f"compare: sources {a!r} {b!r} require explicit allow_broadcast_all_sources=True")
        elif allow_broadcast_all_sources is True:
            pass
        else:
            raise TypeError(f"invalid type for allow_broadcast_all_sources: {type(allow_broadcast_all_sources)}")
    if dim_order:
        all_dims.sort(key=lambda d: dim_order.index(d) if d in dim_order else len(dim_order))
    out = Tensor(name, dims=all_dims, dtype=src_dtype)
    out.feature_dim = res_feature_dim(a, b)
    if copy_sparse_dim:
        out.sparse_dim = res_sparse_dim(a, b)
    if not allow_scalar or a.dims:
        a_raw = a.copy_compatible_to_dims_raw(all_dims)
    else:
        a_raw = a.raw_tensor
    if not allow_scalar or b.dims:
        b_raw = b.copy_compatible_to_dims_raw(all_dims)
    else:
        b_raw = b.raw_tensor
    return out, a_raw, b_raw


def res_feature_dim(a: Tensor, b: Tensor) -> Optional[Dim]:
    """
    :param a:
    :param b:
    :return: feature dim if consistent or None
    """
    if a.feature_dim and not b.feature_dim:
        return a.feature_dim
    if b.feature_dim and not a.feature_dim:
        return b.feature_dim
    if a.feature_dim and b.feature_dim and a.feature_dim == b.feature_dim:
        return a.feature_dim
    return None


def res_sparse_dim(a: Tensor, b: Tensor) -> Optional[Dim]:
    """
    :param a:
    :param b:
    :return: sparse dim if consistent or None
    """
    if a.sparse_dim and not b.sparse_dim:
        return a.sparse_dim
    if b.sparse_dim and not a.sparse_dim:
        return b.sparse_dim
    if a.sparse_dim and b.sparse_dim and a.sparse_dim == b.sparse_dim:
        return a.sparse_dim
    return None


def strided_slice_raw_key(
    tensor: Tensor,
    axis: Optional[Union[Dim, Sequence[Dim]]] = None,
    key: Optional[rf.ItemKeyType] = None,
    key_dim: Optional[Union[Dim, Sequence[Dim]]] = None,
) -> Tuple[Union[slice, int, T, Sequence[Union[None, slice, int, T]]], Tuple[Dim, ...]]:
    """
    Given an axis and a key, return a raw key that can be used to index into a raw tensor,
    as in `raw_tensor[key]`.
    The tensor is needed to infer the raw axis index.

    :return: raw item key, resulting dims of reduced tensor
    """
    if key is None or (isinstance(key, (list, tuple)) and len(key) == 0):
        return slice(None, None), tensor.dims
    if isinstance(key, Tensor) and key.dtype == "bool":
        raise NotImplementedError("strided_slice: boolean mask")
    if isinstance(key, (slice, int, numpy.number, numpy.ndarray, Tensor)):
        if axis is None:
            axis = _slice_find_sparse_dim(key)
        if not isinstance(axis, Dim):
            raise TypeError(f"strided_slice: must specify axis for key, got {type(axis).__name__}")
        axis_int = tensor.get_axis_from_description(axis)
        key_raw = _map_slice_value_raw(key)
        if _slice_value_is_reduce(key):
            res_dims = tuple(d for (i, d) in enumerate(tensor.dims) if i != axis_int)
        else:
            if not isinstance(key_dim, Dim):
                raise TypeError(f"strided_slice: expected key_dim of type Dim, got {key_dim}")
            res_dims = tuple(d if i != axis_int else key_dim for (i, d) in enumerate(tensor.dims))
        if axis_int == 0:
            return key_raw, res_dims
        return (None,) * axis_int + (key_raw,), res_dims
    if not isinstance(key, (list, tuple)):
        raise TypeError(f"strided_slice: unexpected key type: {type(key).__name__}")
    if axis is not None:
        if not isinstance(axis, (list, tuple)):
            raise TypeError(
                f"strided_slice: key is sequence, thus expect axis to be sequence as well, got {type(axis).__name__}"
            )
        if len(axis) != len(key):
            raise ValueError(f"strided_slice: mismatching axis seq len {len(axis)} and key seq length {len(key)}")
    if key_dim is not None:
        if not isinstance(key_dim, (list, tuple)):
            raise TypeError(
                f"strided_slice: key is sequence, thus expect key_dim to be sequence as well,"
                f" got {type(key_dim).__name__}"
            )
        if len(key_dim) != len(key):
            raise ValueError(f"strided_slice: mismatching key_dim seq len {len(key_dim)} and key seq length {len(key)}")
    raw_out_keys = {}  # per raw axis
    raw_out_dims: List[Optional[Dim]] = list(tensor.dims)
    for i, key_ in enumerate(key):
        axis_ = None
        if axis is not None:
            axis_ = axis[i]
        if axis_ is None:
            axis_ = _slice_find_sparse_dim(key_)
        if not isinstance(axis_, Dim):
            raise TypeError(f"strided_slice: must specify axis for key sequence, got {type(axis_).__name__}")
        axis_int = tensor.get_axis_from_description(axis_)
        key_raw = _map_slice_value_raw(key_)
        if axis_int in raw_out_keys:
            raise ValueError(f"strided_slice: duplicate axis {axis_} in sequence")
        raw_out_keys[axis_int] = key_raw
        if _slice_value_is_reduce(key_):
            raw_out_dims[axis_int] = None
        else:
            key_dim_ = None
            if key_dim is not None:
                key_dim_ = key_dim[i]
            if not isinstance(key_dim_, Dim):
                raise TypeError(f"strided_slice: expected key_dim for key sequence, got {type(key_dim_).__name__}")
            raw_out_dims[axis_int] = key_dim_
    out = []
    for i in range(0, max(raw_out_keys) + 1):
        out.append(raw_out_keys.get(i))
    return tuple(out), tuple(d for d in raw_out_dims if d is not None)


def _slice_find_sparse_dim(v: Union[Tensor, slice, Any]) -> Optional[Dim]:
    if isinstance(v, Tensor):
        return v.sparse_dim
    if isinstance(v, slice):
        attribs = {k: getattr(v, k) for k in ("start", "stop", "step")}
        tensors = {k: v for (k, v) in attribs.items() if isinstance(v, Tensor)}
        sparse_dims = {k: v.sparse_dim for (k, v) in tensors.items() if v.sparse_dim}
        sparse_dims_ = list(set(sparse_dims.values()))
        if len(sparse_dims_) == 0:
            return None
        if len(sparse_dims_) == 1:
            return sparse_dims_[0]
        raise ValueError(f"strided_slice: multiple different sparse dims in slice {v}: {sparse_dims}")
    return None


def _map_slice_value_raw(
    v: Union[None, slice, int, numpy.number, numpy.ndarray, Tensor[T]]
) -> Union[None, slice, int, numpy.number, T]:
    if v is None:
        return None
    if isinstance(v, slice):
        return slice(_map_slice_value_raw(v.start), _map_slice_value_raw(v.stop), _map_slice_value_raw(v.step))
    if isinstance(v, (int, numpy.number)):
        return v
    if isinstance(v, numpy.ndarray):
        assert v.ndim <= 1, f"strided_slice: expect scalar or vector, got array with shape {v.shape}"
        return v
    if isinstance(v, Tensor):
        assert len(v.dims) <= 1, f"strided_slice: expect scalar or vector, got Tensor with dims {v.dims}"
        return v.raw_tensor
    raise TypeError(f"strided_slice: got unexpected value of type {type(v).__name__}")


def _slice_value_is_reduce(v: Union[None, slice, int, numpy.number, numpy.ndarray, Tensor[T]]) -> bool:
    if v is None:
        return False
    if isinstance(v, slice):
        return False
    if isinstance(v, (int, numpy.number)):
        return True
    if isinstance(v, numpy.ndarray):
        assert v.ndim <= 1, f"strided_slice: expect scalar or vector, got array with shape {v.shape}"
        return v.ndim == 0
    if isinstance(v, Tensor):
        assert len(v.dims) <= 1, f"strided_slice: expect scalar or vector, got Tensor with dims {v.dims}"
        return v.dims == 0
    raise TypeError(f"strided_slice: got unexpected value of type {type(v).__name__}")



"""
Common settings for linters, e.g. pycharm-inspect.py or pylint.py.

Some resources:
https://github.com/google/styleguide
https://google.github.io/styleguide/pyguide.html
https://chromium.googlesource.com/chromium/src/+/master/docs/code_reviews.md#OWNERS-files

Our Python style:
2 space indentation, 120 char line limit, otherwise mostly PEP8,
and follow further common Python conventions (closely follow PyCharm warnings).

We want to require the CI/Travis PyCharm code style checks to be mandatory for passing.

PyCharm is by far the best tool we know for this.
Pylint or Flake8 do not come close.
They report many more false positives, and fail to detect many problems.
Maybe this changes at some point, but that is why we should use PyCharm for now.
This is the script ```pycharm-inspect.py``.

PyCharm is not perfect either, though.
There are some false positives.
Maybe not always due to PyCharms fault but some lib (TensorFlow or NumPy) does strange things.
We should filter these out.

There are multiple ways how we could filter out false positives:

* ``pycharm-inspect``:

  * `inspect_class_blacklist`.
    Will completely remove/skip some type from the report, even not mention it (as ignored).
    Currently not used at all.
    Maybe could have `PyTypeCheckerInspection`, although the spell checking is useful in general.

  * `inspect_class_not_counted`.
    This will globally not count any such warnings as critical.
    It will still report them.
    This should also only rarely be used, and only for warnings which we generally do not care about,
    or which are false positives in >=50% of the cases or so.
    `PyTypeCheckerInspection` is currently here, and should be removed later,
    once there are less false positives on PyCharm side.
    Actually there are also often true positives, so this is a helpful inspection.
    Otherwise there are things which really are not so critical.

  * In `report_inspect_xml`, there is some extra code to filter out some false positives by common patterns.
    E.g. we normally check for `PyArgumentListInspection`,
    but there are some cases for NumPy which produces false warnings.

* `tests/PyCharm.idea/inspectionProfiles/Project_Default.xml`:
  Includes general inspection settings. And e.g. list of valid imports which would not lead to a warning.
  This file rarely needs to be touched.

* `lint_common.ignore_count_for_files`:
  These are mostly old files, which are not fixed yet, but which are also only rarely used, so not so important.
  New code files should normally not be added here.
  (Unless maybe it's stuff for `extern` which we copied from somewhere.)

* In the code itself, at the place of the false positive:

  * You could add some ``# noinspection PyProtectedMember``, ``# noinspection PyBroadException`` or so
    in front of the statement.
    PyCharm can do that automatically. You click on "suppress warning for this statement" or sth like that.
    Try to only use this rarely.
    This is in most cases a valid warning, no false warning, but you are ok to write the code like this.

  * You could do the same for a false warning.
    But really try to avoid that.
    We should not clutter our code with stuff which are just to workaround a false warning
    in some specific PyCharm version (which hopefully will get fixed in some later PyCharm version).
    Better filter it out by some pattern in `pycharm-inspect`.

  * Maybe your Python code is just somewhat unusual,
    and you can potentially just rewrite it slightly in a cleaner way,
    which also will get rid of the warning.
    If that is the case, definitely do it!
    Maybe some explicit ``assert isinstance(...)`` sometimes can help as well.

  * Your line got too long only because of added type information.
    Just add `` # nopep8`` at the very end of the line.
    This should still be rare.
    If you can (almost always), rewrite it in a way that it spans multiple lines.

  * You have some rare case which needs unusual code and PyCharm will show some warning.
    Then it is valid to add ``# noinspection ...``. Or even ``# noqa`` or ``# nopep8``.

* Spelling dict ``spelling.dic``.
  Do not just everything in there. Only things which you consider so common that we should accept them.
  I.e. words (variable names) like ``tmpdir`` should better just be written as ``tmp_dir``,
  or ``argparser`` as ``arg_parser``, instead of adding all such cases to the spelling dict.

"""

import os

_my_dir = os.path.dirname(os.path.abspath(__file__))
_root_dir = os.path.dirname(_my_dir)
assert os.path.exists("%s/rnn.py" % _root_dir)


# Proceed like this: Fix all warnings for some file, then remove it from this list.
# I removed already all files which really should not have warnings (mostly the TF backend + shared files).
ignore_count_for_files = {
    "returnn/util/fsa.py",
    "returnn/util/task_system.py",
    "returnn/datasets/bundle_file.py",
    "returnn/datasets/cached.py",
    "returnn/datasets/normalization_data.py",
    "returnn/datasets/raw_wav.py",
    "returnn/datasets/stereo.py",
    "returnn/tf/layers/segmental_model.py",
    # Copied to RETURNN, still work-in-progress to clean up.
    "returnn/extern/graph_editor/reroute.py",
    "returnn/extern/graph_editor/subgraph.py",
    "returnn/extern/graph_editor/transform.py",
    "returnn/extern/graph_editor/util.py",
    "returnn/extern/official_tf_resnet/resnet_model.py",
    # Ignore some outdated or rarely used tools/demos.
    "tools/collect-orth-symbols.py",
    "tools/debug-plot-search-scores.py",
    "tools/import-blocks-mt-model.py",
    "tools/import-sprint-nn.py",
    "tools/import-t2t-mt-model.py",
    "demos/mdlstm/IAM/create_IAM_dataset.py",
    "demos/mdlstm/IAM/decode.py",
    "demos/mdlstm/artificial/create_test_h5.py",
    "demos/mdlstm/artificial_rgb/create_test_h5.py",
}


def find_all_py_source_files():
    """
    :rtype: list[str]
    """
    # Earlier this was a `glob("%s/*.py" % _root_dir)`. But not anymore, since we have the new package structure.
    src_files = []
    for root, dirs, files in os.walk(_root_dir):
        if root == _root_dir:
            root = ""
        else:
            assert root.startswith(_root_dir + "/")
            root = root[len(_root_dir) + 1 :]  # relative to the root
            root += "/"
        # Ignore tests, or other irrelevant directories.
        if root == "":
            dirs[:] = ["returnn", "demos", "tools"]
        else:
            dirs[:] = sorted(dirs)
            # Ignore extern git submodules.
            dirs[:] = [d for d in dirs if not os.path.exists("%s/%s%s/.git" % (_root_dir, root, d))]
        for file in sorted(files):
            if not file.endswith(".py"):
                continue
            if file == "_setup_info_generated.py":
                continue
            src_files.append(root + file)
    return src_files



"""
Testing returnn.frontend.decoder.transformer.
"""

from __future__ import annotations

import _setup_test_env  # noqa
import torch
from returnn.util.debug import PyTracer, check_py_traces_rf_to_pt_equal
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


def _setup():
    try:
        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ImportError:
        pass


_setup()


def test_llama():
    """
    Test that we can reproduce the Llama model.

    This here is the final complete test.
    There are several other sub-tests:

    - :func:`test_rotary_embedding`
    - :func:`test_rope_causal_self_att`

    Some references for the whole Llama model:
    https://github.com/meta-llama/llama/blob/main/llama/model.py
    https://github.com/meta-llama/llama3/blob/main/llama/model.py
    https://github.com/karpathy/llama2.c/blob/master/model.py
    https://github.com/huggingface/transformers/blob/main/src/transformers/models/llama/modeling_llama.py
    https://github.com/hkproj/pytorch-llama/blob/main/model.py
    https://github.com/likejazz/llama3.np/blob/main/llama3.py
    """
    from returnn.frontend.decoder.transformer import TransformerDecoder, TransformerDecoderLayer, FeedForwardGated
    from returnn.frontend.conversions.hf_llama import import_params_hf_llama_to_rf_transformer_decoder
    from transformers.models.llama.modeling_llama import LlamaDecoderLayer, LlamaForCausalLM, LlamaModel, LlamaConfig

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )

    model_hf = LlamaForCausalLM(config)
    print("HF Model:")
    print(model_hf)
    print("Parameters:")
    num_params = 0
    for k, v in model_hf.named_parameters():
        print(f"{k}: {list(v.shape)} {v.dtype}")
        num_params += v.numel()
    print("Total number of parameters:", num_params)

    rf.select_backend_torch()

    model_dim = Dim(config.hidden_size, name="model")
    model_rf = TransformerDecoder(
        encoder_dim=None,
        vocab_dim=Dim(config.vocab_size, name="vocab"),
        model_dim=model_dim,
        num_layers=config.num_hidden_layers,
        pos_enc=None,
        norm=rf.RMSNorm,
        ff=FeedForwardGated,
        share_embedding=False,
        input_embedding_scale=1.0,
        decoder_layer_opts=dict(self_att=rf.RotaryPosCausalSelfAttention, self_att_opts=dict(with_bias=False)),
        num_heads=config.num_attention_heads,
        dropout=0,
        att_dropout=0,
    )
    print("RF Model:")
    print(model_rf)
    print("Parameters:")
    num_params = 0
    for k, v in model_rf.named_parameters():
        print(f"{k}: {list(v.dims)} {v.dtype}")
        num_params += v.num_elements()
    print("Total number of parameters:", num_params)

    import_params_hf_llama_to_rf_transformer_decoder(model_hf, model_rf)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    in_ = rf.random_uniform([batch_dim, seq_dim], sparse_dim=model_rf.vocab_dim)
    in_.name = "input_labels"

    with PyTracer([TransformerDecoder.__call__, TransformerDecoderLayer.__call__], Tensor) as trace_rf:
        out_rf, _ = model_rf(in_, spatial_dim=seq_dim, state=model_rf.default_initial_state(batch_dims=[batch_dim]))

    mask = rf.sequence_mask([batch_dim, seq_dim])
    with PyTracer([LlamaForCausalLM.forward, LlamaModel.forward, LlamaDecoderLayer.forward], torch.Tensor) as trace_hf:
        out_hf = model_hf(in_.raw_tensor, attention_mask=mask.raw_tensor)

    check_py_traces_rf_to_pt_equal(
        trace_rf.captured_locals,
        trace_hf.captured_locals,
        [
            (
                (TransformerDecoder.__call__, 0, "decoded", 0),
                (LlamaModel.forward, 0, "inputs_embeds", 0),
                (batch_dim, seq_dim, model_dim),
            ),
        ],
    )

    print("Check...")
    assert out_rf.raw_tensor.shape == out_hf.logits.shape
    torch.testing.assert_allclose(out_rf.raw_tensor, out_hf.logits)
    print("  all matched!")


def test_feed_forward_gated():
    from returnn.frontend.decoder.transformer import FeedForwardGated
    from returnn.frontend.conversions.hf_llama import import_params_hf_llama_mlp_to_rf_feed_forward_gated
    from transformers.models.llama.modeling_llama import LlamaMLP, LlamaConfig

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )

    model_hf = LlamaMLP(config)

    rf.select_backend_torch()
    rf.set_random_seed(42)

    model_dim = Dim(config.hidden_size, name="model")
    model_rf = FeedForwardGated(out_dim=model_dim, ff_dim=Dim(config.intermediate_size, name="inter"), dropout=0.0)

    import_params_hf_llama_mlp_to_rf_feed_forward_gated(model_hf, model_rf)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    in_ = rf.random_uniform([batch_dim, seq_dim, model_dim])
    in_.name = "input"

    out_rf = model_rf(in_)
    out_rf = out_rf.copy_transpose((batch_dim, seq_dim, model_dim))

    out_hf = model_hf(in_.raw_tensor)

    print("Check...")
    assert out_rf.raw_tensor.shape == out_hf.shape
    torch.testing.assert_allclose(out_rf.raw_tensor, out_hf)
    print("  all matched!")



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import numpy as np
import numpy.testing
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model, tf_scope


def _setup():
    try:
        import lovely_tensors

        lovely_tensors.monkey_patch()
    except ImportError:
        pass


_setup()


def test_dot_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    key_dim = Dim(7, name="key")
    value_dim = Dim(13, name="value")
    extern_data = TensorDict(
        {
            "q": Tensor("q", [batch_dim, time_dim, key_dim], dtype="float32"),
            "k": Tensor("k", [batch_dim, time_dim, key_dim], dtype="float32"),
            "v": Tensor("v", [batch_dim, time_dim, value_dim], dtype="float32", feature_dim_axis=2),
        }
    )

    class _Net(rf.Module):
        def __call__(self, q: Tensor, k: Tensor, v: Tensor) -> Tensor:
            kv_axis = Dim(None, name=f"kv-axis")
            k, _ = rf.replace_dim(k, in_dim=time_dim, out_dim=kv_axis)
            v, _ = rf.replace_dim(v, in_dim=time_dim, out_dim=kv_axis)
            return rf.dot_attention(q, k, v, axis=kv_axis, key_dim=key_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(q=extern_data["q"], k=extern_data["k"], v=extern_data["v"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, value_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_self_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.SelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""
            return self.self_att(x, axis=axis)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_self_attention_to_pure_torch():
    # test whether the torch and returnn implementation of the mhsa layer are equivalent
    import torch
    import returnn.frontend as rf
    from returnn.tensor import Dim

    # torch.backends.mha.set_fastpath_enabled(False)

    rf.select_backend_torch()
    rf.init_forward_step_run_ctx()
    rf.set_random_seed(1)

    batch_dim = Dim(3, name="batch")
    spatial_dim = Dim(11, name="spatial")
    out_dim = Dim(54, name="out")
    num_heads = 2

    random_opts = {"distribution": "normal", "dtype": "float32"}
    rf_input = rf.random(dims=[batch_dim, spatial_dim, out_dim], **random_opts)
    qkv_weight = rf.random(dims=[out_dim, 3 * out_dim], **random_opts)
    qkv_bias = rf.random(dims=[3 * out_dim], **random_opts)
    proj_weight = rf.random(dims=[out_dim.copy(match_priority=1), out_dim], **random_opts)
    proj_bias = rf.random(dims=[out_dim], **random_opts)

    rf_mhsa = rf.SelfAttention(
        in_dim=out_dim,
        proj_dim=out_dim,
        key_dim_total=out_dim,
        value_dim_total=out_dim,
        num_heads=num_heads,
    )
    rf_mhsa.qkv.weight.initial = qkv_weight
    rf_mhsa.qkv.bias.initial = qkv_bias
    rf_mhsa.proj.weight.initial = proj_weight
    rf_mhsa.proj.bias.initial = proj_bias

    torch_input = rf_input.raw_tensor
    torch_mhsa = torch.nn.MultiheadAttention(
        out_dim.dimension,
        num_heads,
        batch_first=True,
    )
    torch_mhsa.load_state_dict(
        {
            "in_proj_weight": qkv_weight.raw_tensor.reshape(
                out_dim.dimension, num_heads, 3, out_dim.dimension // num_heads
            )
            .permute(2, 1, 3, 0)
            .reshape(-1, out_dim.dimension),
            "in_proj_bias": qkv_bias.raw_tensor.reshape(num_heads, 3, out_dim.dimension // num_heads)
            .permute(1, 0, 2)
            .reshape(-1),
            "out_proj.weight": proj_weight.raw_tensor.reshape(
                num_heads, out_dim.dimension // num_heads, out_dim.dimension
            )
            .permute(2, 0, 1)
            .reshape(-1, out_dim.dimension),
            "out_proj.bias": proj_bias.raw_tensor,
        }
    )
    torch_mhsa.eval()

    rf_output = rf_mhsa(rf_input, axis=spatial_dim)
    torch_output, torch_attn_weights = torch_mhsa(torch_input, torch_input, torch_input, key_padding_mask=None)

    print("RF output")
    print(rf_output.raw_tensor)
    print(rf_output.raw_tensor.shape)
    print("---------------------------")
    print("Torch output")
    print(torch_output)
    print(torch_output.shape)

    torch.testing.assert_allclose(rf_output.raw_tensor, torch_output, atol=1e-3, rtol=1e-4)


def test_causal_self_attention():
    from returnn.tensor import single_step_dim

    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.CausalSelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""

            def _body(_x: Tensor, _state: rf.State) -> Tuple[Tensor, rf.State]:
                _y, _state.self_att = self.self_att(_x, axis=single_step_dim, state=_state.self_att)
                return _y, _state

            y, _, _ = rf.scan(
                spatial_dim=axis,
                xs=x,
                body=_body,
                ys=Tensor("y", dims=[batch_dim, self.out_dim], dtype="float32"),
                initial=rf.State(self_att=self.self_att.default_initial_state(batch_dims=[batch_dim])),
            )
            return y

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        # TF needs TensorArray unstack, not implemented yet
        test_tensorflow=False,
    )


def test_rotary_embedding():
    import torch
    from transformers.models.llama.modeling_llama import LlamaConfig, LlamaRotaryEmbedding

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )
    model_hf = LlamaRotaryEmbedding(config=config)

    rf.select_backend_torch()
    rf.set_random_seed(42)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    model_head_dim = Dim(config.hidden_size // config.num_attention_heads, name="model_head")

    # base is a bit different in rf.sinusoidal_positional_encoding (like the original)
    # vs how it's used for RoPE.
    # log(base) / (dim / 2 - 1) = log(10_000) * 2 / dim
    # <=> log(base) = log(10_000) * (dim / 2 - 1) * 2 / dim = log(10_000) * (1 - 2 / dim)
    # <=> base = 10_000 ** (1 - 2 / dim)
    out_rf = rf.sinusoidal_positional_encoding(
        spatial_dim=seq_dim, feat_dim=model_head_dim, base=10_000 ** (1 - 2 / model_head_dim.dimension)
    )
    print("out_rf:", out_rf)
    # For the comparison below.
    out_rf = rf.expand_dim(out_rf, batch_dim)
    out_rf = out_rf.copy_transpose((batch_dim, seq_dim, model_head_dim))
    # Split the sin/cos for the comparison below.
    out_rf_sin, out_rf_cos = rf.split(out_rf, axis=model_head_dim, out_dims=[model_head_dim.div_left(2)] * 2)
    print("out_rf':", out_rf_sin, out_rf_cos)

    position_ids = rf.expand_dim(rf.range_over_dim(seq_dim), batch_dim)  # LlamaRotaryEmbedding wants this
    out_hf_cos, out_hf_sin = model_hf(
        torch.zeros(()), position_ids=position_ids.copy_compatible_to_dims_raw((batch_dim, seq_dim))
    )
    print("out_hf:", out_hf_sin, out_hf_cos)
    # The values are just repeated. Cut off the second half for comparison.
    out_hf_cos = out_hf_cos[:, :, : out_hf_cos.shape[-1] // 2]
    out_hf_sin = out_hf_sin[:, :, : out_hf_sin.shape[-1] // 2]
    print("out_hf':", out_hf_sin, out_hf_cos)

    assert out_rf_sin.raw_tensor.shape == out_hf_sin.shape
    assert out_rf_cos.raw_tensor.shape == out_hf_cos.shape
    torch.testing.assert_allclose(out_rf_sin.raw_tensor, out_hf_sin)
    torch.testing.assert_allclose(out_rf_cos.raw_tensor, out_hf_cos)


def test_rope_causal_self_att():
    import torch
    from returnn.util.pprint import pprint
    from returnn.util.debug import PyTracer, check_py_traces_rf_to_pt_equal

    # noinspection PyProtectedMember
    from returnn.frontend.attention import _apply_rope as rf_apply_rope
    from returnn.frontend.conversions.hf_llama import import_params_hf_llama_att_to_rf_rotary_att

    from transformers.models.llama.modeling_llama import (
        LlamaAttention,
        LlamaRotaryEmbedding,
        LlamaConfig,
        apply_rotary_pos_emb,
    )

    config = LlamaConfig(
        vocab_size=11,
        hidden_size=64,
        intermediate_size=64 * 4,
        num_hidden_layers=2,
        num_attention_heads=2,
        max_position_embeddings=128,
    )

    model_hf = LlamaAttention(config, layer_idx=0)

    rf.select_backend_torch()
    rf.set_random_seed(42)

    model_dim = Dim(config.hidden_size, name="model")
    model_rf = rf.RotaryPosCausalSelfAttention(
        in_dim=model_dim,
        proj_dim=model_dim,
        num_heads=config.num_attention_heads,
        key_dim_total=model_dim,
        value_dim_total=model_dim,
        with_bias=False,
    )
    import_params_hf_llama_att_to_rf_rotary_att(model_hf, model_rf)

    batch_dim = Dim(3, name="batch")
    seq_dim = Dim(rf.random_uniform([batch_dim], minval=7, maxval=13, dtype="int32"), name="seq")
    in_ = rf.random_uniform([batch_dim, seq_dim, model_dim])
    in_.name = "input"

    with PyTracer(
        [rf.RotaryPosCausalSelfAttention.__call__, rf.sinusoidal_positional_encoding, rf.dot_attention, rf_apply_rope],
        (Tensor, Dim),
    ) as trace_rf:
        out_rf, _ = model_rf(in_, axis=seq_dim, state=model_rf.default_initial_state(batch_dims=[batch_dim]))
        out_rf = out_rf.copy_transpose((batch_dim, seq_dim, model_dim))
    pprint(trace_rf.captured_locals)

    position_ids = rf.expand_dim(rf.range_over_dim(seq_dim), batch_dim)  # LlamaRotaryEmbedding wants this
    with PyTracer(
        [LlamaAttention.forward, LlamaRotaryEmbedding.forward, apply_rotary_pos_emb], torch.Tensor
    ) as trace_hf:
        # causal_mask code copied from LlamaAttention
        sequence_length = target_length = in_.raw_tensor.shape[1]
        causal_mask = torch.full(
            (sequence_length, target_length),
            fill_value=torch.finfo(in_.raw_tensor.dtype).min,
            dtype=in_.raw_tensor.dtype,
            device=in_.raw_tensor.device,
        )
        causal_mask = torch.triu(causal_mask, diagonal=1)
        cache_position = torch.arange(0, in_.raw_tensor.shape[1], device=in_.raw_tensor.device)
        causal_mask *= torch.arange(target_length, device=cache_position.device) > cache_position.reshape(-1, 1)
        causal_mask = causal_mask[None, None, :, :].expand(in_.raw_tensor.shape[0], 1, -1, -1)
        rotary_emb = LlamaRotaryEmbedding(config=config)
        position_embeddings = rotary_emb(
            torch.zeros(()), position_ids=position_ids.copy_compatible_to_dims_raw((batch_dim, seq_dim))
        )
        out_hf, _, _ = model_hf(in_.raw_tensor, attention_mask=causal_mask, position_embeddings=position_embeddings)
    pprint(trace_hf.captured_locals)

    print("First HF att weight tensor:")
    print(trace_hf.captured_locals[LlamaAttention.forward][0]["attn_weights"][2][0, 0, 0].detach().numpy())

    check_py_traces_rf_to_pt_equal(
        trace_rf.captured_locals,
        trace_hf.captured_locals,
        [
            (
                (rf.RotaryPosCausalSelfAttention.__call__, 0, "q", 0),
                (LlamaAttention.forward, 0, "query_states", 1),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x.reshape(*x.shape[:-1], 2, -1).transpose(-1, -2).flatten(-2),
                    dims=(batch_dim, model_rf.num_heads, seq_dim, model_rf.key_dim_per_head),
                    name=name,
                ),
            ),
            (
                (rf.RotaryPosCausalSelfAttention.__call__, 0, "k", 0),
                (LlamaAttention.forward, 0, "key_states", 1),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x.reshape(*x.shape[:-1], 2, -1).transpose(-1, -2).flatten(-2),
                    dims=(batch_dim, model_rf.num_heads, seq_dim, model_rf.key_dim_per_head),
                    name=name,
                ),
            ),
            (
                (rf.sinusoidal_positional_encoding, 0, "div_term", 0),
                (LlamaRotaryEmbedding.forward, 0, "inv_freq_expanded", 0),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x[0, :, 0], dims=[model_rf.key_dim_per_head.div_left(2)], name=name
                ),
            ),
            (
                (rf.sinusoidal_positional_encoding, 0, "arg_sin", 0),
                (LlamaRotaryEmbedding.forward, 0, "freqs", 0),
                lambda x, *, name, resolve_dim, **_: rf.convert_to_tensor(
                    x[0],
                    dims=(resolve_dim("spatial_dim"), model_rf.key_dim_per_head.div_left(2)),
                    name=name,
                ),
            ),
            (
                (rf_apply_rope, 0, "pe_imag", 0),
                (apply_rotary_pos_emb, 0, "sin", 0),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x[0, :, : x.shape[2] // 2], dims=(seq_dim, model_rf.key_dim_per_head.div_left(2)), name=name
                ),
            ),
            (
                (rf_apply_rope, 0, "pe_imag", 0),
                (apply_rotary_pos_emb, 0, "sin", 0),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x[-1, :, x.shape[2] // 2 :], dims=(seq_dim, model_rf.key_dim_per_head.div_left(2)), name=name
                ),
            ),
            (
                (rf_apply_rope, 0, "pe_real", 0),
                (apply_rotary_pos_emb, 0, "cos", 0),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x[0, :, : x.shape[2] // 2],
                    dims=(seq_dim, model_rf.key_dim_per_head.div_left(2)),
                    name=name,
                ),
            ),
            (
                (rf.RotaryPosCausalSelfAttention.__call__, 0, "q", -1),
                (LlamaAttention.forward, 0, "query_states", -1),
                lambda x, *, name, **_: rf.convert_to_tensor(
                    x.reshape(*x.shape[:-1], 2, -1).transpose(-1, -2).flatten(-2),
                    dims=(batch_dim, model_rf.num_heads, seq_dim, model_rf.key_dim_per_head),
                    name=name,
                ),
            ),
            (
                (rf.dot_attention, 0, "energy", 0),
                (LlamaAttention.forward, 0, "attn_weights", 0),
                (batch_dim, model_rf.num_heads, seq_dim, "axis"),
            ),
            (
                (rf.dot_attention, 0, "att_weights", 0),
                (LlamaAttention.forward, 0, "attn_weights", 2),
                (batch_dim, model_rf.num_heads, seq_dim, "axis"),
            ),
            (
                (rf.dot_attention, 0, "att", 0),
                (LlamaAttention.forward, 0, "attn_output", 0),
                (batch_dim, model_rf.num_heads, seq_dim, model_rf.value_dim_per_head),
            ),
        ],
    )

    print("Final check...")
    assert out_rf.raw_tensor.shape == out_hf.shape
    torch.testing.assert_allclose(out_rf.raw_tensor, out_hf)
    print("  all matched!")


def test_relative_positional_encoding():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(8, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor, *, axis: Dim) -> Tuple[Tensor, Dim]:
            x, dim = rf.relative_positional_encoding(
                key_value_spatial_dim=axis, query_spatial_dim=axis, feat_dim=in_dim
            )
            return x, dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, dim = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_rel_pos_self_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(8, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.RelPosSelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""
            return self.self_att(x, axis=axis)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_sinusoidal_positional_encoding():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(8, name="feat")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, feat_dim], dtype="float32"),
        }
    )

    def _forward_step(**_kwargs):
        out = rf.sinusoidal_positional_encoding(spatial_dim=time_dim, feat_dim=feat_dim)
        out.mark_as_default_output(shape=(time_dim, feat_dim))

    res = run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)

    from returnn.tf.util import basic as tf_util

    with tf_scope() as session:
        tf_ref = tf_util.get_positional_encoding(
            num_channels=feat_dim.dimension, length=res.data["output"].raw_tensor.shape[0]
        )
        tf_ref_v = session.run(tf_ref)

    np.testing.assert_almost_equal(res.data["output"].raw_tensor, tf_ref_v, decimal=5)


def test_CausalSelfAttention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(8, name="feat")
    key_dim = Dim(6, name="key")
    value_dim = Dim(10, name="value")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, feat_dim], dtype="float32"),
        }
    )

    def _forward_step(*, model: rf.CausalSelfAttention, extern_data: TensorDict):
        data = extern_data["data"]
        data.mark_as_output("data", shape=[batch_dim, time_dim, feat_dim])
        time_dim.dyn_size_ext.mark_as_output("seq_len", shape=[batch_dim])
        out, _ = model(data, axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, value_dim))
        model.qkv.weight.mark_as_output("qkv_weight", shape=[feat_dim, 2 * key_dim + value_dim])

    res = run_model(
        extern_data,
        lambda *, epoch, step: rf.CausalSelfAttention(
            in_dim=feat_dim,
            proj_dim=None,
            key_dim_total=key_dim,
            value_dim_total=value_dim,
            num_heads=2,
            with_bias=False,
        ),
        _forward_step,
        # Some problem with dimension tags currently in the TF-layers-dict backend...
        # Anyway, we compare to the TF SelfAttentionLayer with attention_left_only=True below.
        test_tensorflow=False,
    )

    extern_data.reset_content()

    with tf_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData

        net_dict = {
            "self_att": {
                "class": "self_attention",
                "from": "data",
                "num_heads": 2,
                "total_key_dim": key_dim.dimension,
                "attention_left_only": True,
                "out_dim": value_dim,
                "is_output_layer": True,
            }
        }
        net = TFNetwork(
            extern_data=ExternData(
                {
                    "data": {
                        "dims": [batch_dim, time_dim, feat_dim],
                        "time_dim_axis": 1,
                        "feature_dim_axis": 2,
                        "dtype": "float32",
                        "version": 1,
                    }
                }
            )
        )
        net.construct_from_dict(net_dict)
        layer = net.get_default_output_layer()
        layer.params["QKV"].load(res.data["qkv_weight"].raw_tensor, session=session)
        out = layer.output.copy_transpose([batch_dim, time_dim, value_dim]).copy_masked(0.0)

        out_tf_v = session.run(
            out.raw_tensor,
            feed_dict={
                net.extern_data.data["data"].placeholder: res.data["data"].raw_tensor,
                net.extern_data.data["data"].dims[1].dyn_size_ext.raw_tensor: res.data["seq_len"].raw_tensor,
            },
        )
        numpy.testing.assert_almost_equal(res.data["output"].raw_tensor, out_tf_v, decimal=5)



# -*- coding: utf8 -*-

import sys
import os

import _setup_test_env  # noqa

sys.path += ["tools"]

from hdf_dump import *
import os
from returnn.log import log
import tempfile
from returnn.datasets.generating import DummyDataset
from returnn.datasets.hdf import HDFDataset
from returnn.util.basic import DictAsObj
import unittest
from returnn.util import better_exchook


options = {"epoch": 1, "start_seq": 0, "end_seq": float("inf")}


def test_hdf_dataset_init():
    hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-init")
    hdf_dataset_init(hdf_filename)
    assert os.path.exists(hdf_filename)
    os.remove(hdf_filename)


def test_hdf_create():
    hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-create")
    hdf_dataset = hdf_dataset_init(hdf_filename)
    assert os.path.exists(hdf_filename)

    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
    dataset.init_seq_order(epoch=1)

    hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
    hdf_close(hdf_dataset)

    os.remove(hdf_filename)


def test_hdf_create_and_load():
    hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-load")
    hdf_dataset = hdf_dataset_init(hdf_filename)
    assert os.path.exists(hdf_filename)

    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
    dataset.init_seq_order(epoch=1)

    hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
    hdf_close(hdf_dataset)

    loaded_dataset = HDFDataset()
    loaded_dataset.add_file(hdf_filename)

    os.remove(hdf_filename)


def test_hdf_create_unicode_labels():
    hdf_filename = tempfile.mktemp(suffix=".hdf", prefix="nose-dataset-create")
    hdf_dataset = hdf_dataset_init(hdf_filename)
    assert os.path.exists(hdf_filename)

    dataset = DummyDataset(input_dim=2, output_dim=3, num_seqs=4)
    assert "classes" in dataset.get_target_list()
    dataset.labels["classes"] = ["’", "ä", "x"]  # have some Unicode chars here
    dataset.init_seq_order(epoch=1)

    hdf_dump_from_dataset(dataset, hdf_dataset, DictAsObj(options))
    hdf_close(hdf_dataset)

    os.remove(hdf_filename)


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



#!/usr/bin/env python3

"""
Creates a HDF file, which can be read by :class:`HDFDataset`.
The input is any other dataset (:class:`Dataset`).
"""

from __future__ import annotations

import sys
import argparse

import _setup_returnn_env  # noqa
from returnn.log import log
import returnn.__main__ as rnn
import returnn.datasets.hdf as hdf_dataset_mod
from returnn.datasets import Dataset, init_dataset
from returnn.config import Config


def hdf_dataset_init(file_name):
    """
    :param str file_name: filename of hdf dataset file in the filesystem
    :rtype: hdf_dataset_mod.HDFDatasetWriter
    """
    return hdf_dataset_mod.HDFDatasetWriter(filename=file_name)


def hdf_dump_from_dataset(dataset, hdf_dataset, parser_args):
    """
    :param Dataset dataset: could be any dataset implemented as child of Dataset
    :param hdf_dataset_mod.HDFDatasetWriter hdf_dataset:
    :param parser_args: argparse object from main()
    """
    hdf_dataset.dump_from_dataset(
        dataset=dataset,
        epoch=parser_args.epoch,
        start_seq=parser_args.start_seq,
        end_seq=parser_args.end_seq,
        use_progress_bar=True,
    )


def hdf_close(hdf_dataset):
    """
    :param HDFDataset.HDFDatasetWriter hdf_dataset: to close
    """
    hdf_dataset.close()


def init(config_filename, cmd_line_opts, dataset_config_str):
    """
    :param str config_filename: global config for CRNN
    :param list[str] cmd_line_opts: options for init_config method
    :param str dataset_config_str: dataset via init_dataset_via_str()
    """
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    if config_filename:
        rnn.init_config(config_filename, cmd_line_opts)
        rnn.init_log()
    else:
        log.initialize(verbosity=[5])
    print("Returnn hdf_dump starting up.", file=log.v3)
    rnn.init_faulthandler()
    if config_filename:
        rnn.init_data()
        rnn.print_task_properties()
        assert isinstance(rnn.train_data, Dataset)
        dataset = rnn.train_data
    else:
        assert dataset_config_str
        dataset = init_dataset(dataset_config_str)
    print("Source dataset:", dataset.len_info(fast=True), file=log.v3)
    return dataset


def _is_crnn_config(filename):
    """
    :param str filename:
    :rtype: bool
    """
    if filename.endswith(".gz"):
        return False
    if filename.endswith(".config"):
        return True
    # noinspection PyBroadException
    try:
        config = Config()
        config.load_file(filename)
        return True
    except Exception:
        pass
    return False


def main(argv):
    """
    Main entry.
    """
    parser = argparse.ArgumentParser(description="Dump dataset or subset of dataset into external HDF dataset")
    parser.add_argument(
        "config_file_or_dataset", type=str, help="Config file for RETURNN, or directly the dataset init string"
    )
    parser.add_argument("hdf_filename", type=str, help="File name of the HDF dataset, which will be created")
    parser.add_argument("--start_seq", type=int, default=0, help="Start sequence index of the dataset to dump")
    parser.add_argument("--end_seq", type=int, default=float("inf"), help="End sequence index of the dataset to dump")
    parser.add_argument("--epoch", type=int, default=1, help="Optional start epoch for initialization")

    args = parser.parse_args(argv[1:])
    returnn_config = None
    dataset_config_str = None
    if _is_crnn_config(args.config_file_or_dataset):
        returnn_config = args.config_file_or_dataset
    else:
        dataset_config_str = args.config_file_or_dataset
    dataset = init(config_filename=returnn_config, cmd_line_opts=[], dataset_config_str=dataset_config_str)
    hdf_dataset = hdf_dataset_init(args.hdf_filename)
    hdf_dump_from_dataset(dataset, hdf_dataset, args)
    hdf_close(hdf_dataset)

    rnn.finalize()


if __name__ == "__main__":
    main(sys.argv)
